#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2022 fjord-technologies
# SPDX-License-Identifier: GPL-3.0-or-later
"""fd-openvpn"""

import errno
import sys
import os

import logging
from logging.handlers import WatchedFileHandler

from six.moves.urllib.parse import urlparse

import requests

import yaml
try:
    from yaml import CSafeLoader as YamlLoader, CSafeDumper as YamlDumper
except ImportError:
    from yaml import SafeLoader as YamlLoader, SafeDumper as YamlDumper


OPENVPN_PLUGIN_FUNC_SUCCESS = 0
OPENVPN_PLUGIN_FUNC_ERROR = 1
OPENVPN_PLUGIN_FUNC_DEFERRED = 2

SYSLOG_NAME             = "fd-openvpn"
LOG                     = logging.getLogger(SYSLOG_NAME)

DEFAULT_CONFFILE        = "/etc/fd-vpn/fd-openvpn.yml"
DEFAULT_LOGFILE         = "/var/log/fd-vpn/fd-openvpn.log"
DEFAULT_REQUEST_TIMEOUT = 15

FD_OPENVPN_CONFFILE = os.environ.get('FD_OPENVPN_CONFFILE') or DEFAULT_CONFFILE
FD_OPENVPN_LOGFILE = os.environ.get('FD_OPENVPN_LOGFILE') or DEFAULT_LOGFILE
FD_OPENVPN_LOGLEVEL = getattr(logging, (os.environ.get('FD_OPENVPN_LOGLEVEL') or 'info').upper(), logging.INFO)
FD_OPENVPN_REQUEST_TIMEOUT = os.environ.get('FD_OPENVPN_REQUEST_TIMEOUT') or DEFAULT_REQUEST_TIMEOUT


_CONFIG  = {'fd_vpn_url': 'http://127.0.0.1:8670',
            'request_timeout': FD_OPENVPN_REQUEST_TIMEOUT}

_OPTIONS = {'loglevel': FD_OPENVPN_LOGLEVEL,
            'conffile': FD_OPENVPN_CONFFILE,
            'logfile': FD_OPENVPN_LOGFILE}


def init_logger(logfile, name = None): # pylint: disable=unused-argument
    xformat     = "%(levelname)s:%(asctime)-15s %(name)s[%(process)d][%(threadName)s]: %(message)s"
    datefmt     = '%Y-%m-%d %H:%M:%S'
    logging.basicConfig(level   = logging.DEBUG,
                        format  = xformat,
                        datefmt = datefmt)
    filehandler = WatchedFileHandler(logfile)
    filehandler.setFormatter(logging.Formatter(xformat,
                                               datefmt=datefmt))
    root_logger = logging.getLogger('')
    root_logger.addHandler(filehandler)

    return root_logger

def load_yaml(stream, Loader = YamlLoader):
    return yaml.load(stream, Loader = Loader)

def load_conf_yaml_file(filepath, config_dir = None):
    if config_dir and not filepath.startswith(os.path.sep):
        filepath = os.path.join(config_dir, filepath)

    if not os.path.isfile(filepath):
        return {}

    with open(filepath, 'r') as f:
        return load_yaml(f)

def make_logdir(logfile):
    logdir = os.path.dirname(logfile)
    if not logdir or os.path.exists(logdir):
        return

    try:
        os.makedirs(logdir)
    except OSError as e:
        if e.errno != errno.EEXIST or not os.path.isdir(logdir):
            raise

class FdOpenVpnPlugin(object):
    def __init__(self, *args):
        self.envp = self.unpack_argv()
        self._uri = urlparse(_CONFIG['fd_vpn_url'])
        self._request_timeout = _CONFIG['request_timeout']

    @staticmethod
    def unpack_argv():
        return dict((k, v) for k, v in [x.split('=', 1) for x in sys.argv])

    @staticmethod
    def _build_uri(uri, path = None):
        r = uri
        if path:
            r = r._replace(path = path)

        return r.geturl()

    def _request(self, uri, path = None, method = 'get', params = None):
        req = None

        try:
            url = self._build_uri(uri, path)
            LOG.info("request uri %s %r", method.upper(), url)
            req = getattr(requests, method)(url,
                                            timeout = self._request_timeout,
                                            json = params)
            req.raise_for_status()

            if not req.text:
                return None

            try:
                return req.json()
            except ValueError:
                req.raise_for_status()
        except Exception as e:
            LOG.exception(e)
            raise
        finally:
            if req:
                req.close()

        return False

    def up(self):
        LOG.info("UP")
        LOG.debug(self.envp)

        return OPENVPN_PLUGIN_FUNC_SUCCESS

    def down(self):
        LOG.info("DOWN")
        LOG.debug(self.envp)

        return OPENVPN_PLUGIN_FUNC_SUCCESS

    def route_up(self):
        LOG.info("ROUTE UP")
        LOG.debug(self.envp)

        return OPENVPN_PLUGIN_FUNC_SUCCESS

    def ipchange(self):
        LOG.info("IPCHANGE")
        LOG.debug(self.envp)

        return OPENVPN_PLUGIN_FUNC_SUCCESS

    def tls_verify(self):
        LOG.info("TLS VERIFY")
        LOG.debug(self.envp)

        return OPENVPN_PLUGIN_FUNC_SUCCESS

    def auth_user_pass_verify(self):
        LOG.info("AUTH USER PASS VERIFY")
        LOG.debug(self.envp)
        r = self._request(self._uri,
                          "api/vpn_client/authenticate/%s" % self.envp['username'],
                          'post',
                          params = {'password': self.envp.get('password')})
        if not r:
            return OPENVPN_PLUGIN_FUNC_ERROR

        return OPENVPN_PLUGIN_FUNC_SUCCESS

    def client_connect(self):
        LOG.info("CLIENT CONNECT")
        LOG.debug(self.envp)
        reference = self.envp['common_name'].split('.', 1)[0]
        r = self._request(self._uri,
                          "api/vpn_client/connect/%s" % reference,
                          'patch')
        if not r:
            return OPENVPN_PLUGIN_FUNC_ERROR

        return OPENVPN_PLUGIN_FUNC_SUCCESS

    def client_disconnect(self):
        LOG.info("CLIENT DISCONNECT")
        LOG.debug(self.envp)

        reference = self.envp['common_name'].split('.', 1)[0]
        r = self._request(self._uri,
                          "api/vpn_client/disconnect/%s" % reference,
                          'patch')
        if not r:
            return OPENVPN_PLUGIN_FUNC_ERROR

        return OPENVPN_PLUGIN_FUNC_SUCCESS

    def learn_address(self):
        LOG.info("LEARN ADDRESS")
        LOG.debug(self.envp)

        return OPENVPN_PLUGIN_FUNC_SUCCESS

    def client_connect_v2(self):
        LOG.info("CLIENT CONNECT V2")
        LOG.debug(self.envp)
        reference = self.envp['common_name'].split('.', 1)[0]
        r = self._request(self._uri,
                          "api/vpn_client/connect/%s" % reference,
                          'patch')
        if not r:
            return OPENVPN_PLUGIN_FUNC_ERROR

        return OPENVPN_PLUGIN_FUNC_SUCCESS

    def tls_final(self):
        LOG.info("TLS FINAL")
        LOG.debug(self.envp)

        return OPENVPN_PLUGIN_FUNC_SUCCESS

    def enable_pf(self):
        LOG.info("ENABLE PF")
        LOG.debug(self.envp)

        return OPENVPN_PLUGIN_FUNC_ERROR

    def route_predown(self):
        LOG.info("ROUTE PREDOWN")
        LOG.debug(self.envp)

        return OPENVPN_PLUGIN_FUNC_SUCCESS

    def n(self):
        LOG.info("N")
        LOG.debug(self.envp)

        return OPENVPN_PLUGIN_FUNC_SUCCESS

    def unknown_plugin_type(self):
        LOG.info("UNKNOWN PLUGIN TYPE")
        LOG.debug(self.envp)

        return OPENVPN_PLUGIN_FUNC_SUCCESS


def OPENVPN_PLUGIN_UP(*args):
    return FdOpenVpnPlugin(*args).up()

def OPENVPN_PLUGIN_DOWN(*args):
    return FdOpenVpnPlugin(*args).down()

def OPENVPN_PLUGIN_ROUTE_UP(*args):
    return FdOpenVpnPlugin(*args).route_up()

def OPENVPN_PLUGIN_IPCHANGE(*args):
    return FdOpenVpnPlugin(*args).ipchange()

def OPENVPN_PLUGIN_TLS_VERIFY(*args):
    return FdOpenVpnPlugin(*args).tls_verify()

def OPENVPN_PLUGIN_AUTH_USER_PASS_VERIFY(*args):
    return FdOpenVpnPlugin(*args).auth_user_pass_verify()

def OPENVPN_PLUGIN_CLIENT_CONNECT(*args):
    return FdOpenVpnPlugin(*args).client_connect()

def OPENVPN_PLUGIN_CLIENT_DISCONNECT(*args):
    return FdOpenVpnPlugin(*args).client_disconnect()

def OPENVPN_PLUGIN_LEARN_ADDRESS(*args):
    return FdOpenVpnPlugin(*args).learn_address()

def OPENVPN_PLUGIN_CLIENT_CONNECT_V2(*args):
    return FdOpenVpnPlugin(*args).client_connect_v2()

def OPENVPN_PLUGIN_TLS_FINAL(*args):
    return FdOpenVpnPlugin(*args).tls_final()

def OPENVPN_PLUGIN_ENABLE_PF(*args):
    return FdOpenVpnPlugin(*args).enable_pf()

def OPENVPN_PLUGIN_ROUTE_PREDOWN(*args):
    return FdOpenVpnPlugin(*args).route_predown()

def OPENVPN_PLUGIN_N(*args):
    return FdOpenVpnPlugin(*args).n()

def OPENVPN_UNKNOWN_PLUGIN_TYPE(*args):
    return FdOpenVpnPlugin(*args).unknown_plugin_type()


def main(options):
    """
    Main function; start the server
    """
    global _CONFIG

    make_logdir(options['logfile'])

    root_logger = init_logger(options['logfile'], SYSLOG_NAME)
    _CONFIG.update(load_conf_yaml_file(options['conffile']))

    try:
        root_logger.setLevel(options['loglevel'])
        os.umask(0o22)
    except (KeyboardInterrupt, SystemExit):
        pass
    except Exception:
        LOG.exception("bad things happen")

if __name__ == 'fd_openvpn.plugin':
    main(_OPTIONS)
