# -*- coding: utf-8 -*-
"""vpn_client module"""

import logging
import os
import base64
import re
import shlex
import shutil
import subprocess
import tempfile
import time
import uuid
import zipfile

import bcrypt
import qrcode
import pyotp

from six import BytesIO, ensure_str, iteritems, itervalues

from fd_vpn.objects.vpn_client import FdVpnObjectVPNClient
from fd_vpn.objects.vpn_server import FdVpnObjectVPNServer
from fd_vpn.objects.vpn_user import FdVpnObjectVPNUser

from dwho.classes.modules import DWhoModuleSQLBase, DWhoModuleWebBase, MODULES
from dwho.classes.plugins import PLUGINS
from OpenSSL import crypto
from sonicprobe import helpers
from sonicprobe.libs import xys
from sonicprobe.libs.gencert import GenCert
from sonicprobe.libs.moresynchro import RWLock
from httpdis.ext.httpdis_json import HttpResponse, HttpReqError, HttpReqErrJson


LOG                 = logging.getLogger('fd-vpn.modules.vpn_client')
OPENVPN_EXE_NAME    = 'openvpn-installer.exe'

xys.add_list('vpn_client.auth_method', FdVpnObjectVPNClient.get_auth_method_list())
xys.add_list('vpn_client.sort_by', FdVpnObjectVPNClient.get_columns() + ['vpn_user.username'])
xys.add_list('vpn_client.order_by', ('asc', 'desc'))

xys.add_regex('vpn_client.reference', re.compile(r'^[a-zA-Z0-9]{8}$').match)


class VPNClientModule(DWhoModuleSQLBase, DWhoModuleWebBase):
    MODULE_NAME = 'vpn_client'

    LOCK        = RWLock()

    def __init__(self):
        self.objects    = {'vpn_client': FdVpnObjectVPNClient(),
                           'vpn_server': FdVpnObjectVPNServer(),
                           'vpn_user':   FdVpnObjectVPNUser()}

        DWhoModuleSQLBase.__init__(self)
        DWhoModuleWebBase.__init__(self)

    def init(self, config):
        DWhoModuleSQLBase.init(self, config)
        DWhoModuleWebBase.init(self, config)

        self.objects['vpn_client'].init(config)
        self.objects['vpn_server'].init(config)
        self.objects['vpn_user'].init(config)

    def safe_init(self, options):
        self.lock_timeout       = self.config['general']['lock_timeout']
        self.base_path          = self.config['vpn']['client']['base_path']
        self.ca_name            = self.config['vpn']['client']['ca_name']
        self.common_name        = self.config['vpn']['client']['common_name']

        self.ca_path            = os.path.join(self.base_path, 'ca')
        self.client_path        = os.path.join(self.base_path, 'client')

        self.plugin_iptables    = PLUGINS['iptables']
        self.plugin_openvpn     = PLUGINS['openvpn']
        self.plugin_credentials = PLUGINS['credentials']

    def _get(self, xid, secure = True):
        r = self.objects['vpn_client'].get(to_dict = True,
                                           **{'id': xid})
        if secure:
            r.pop('auth_data', None)

        return r

    def _get_by_reference(self, reference, secure = True):
        r = self.objects['vpn_client'].get(to_dict = True,
                                           **{'reference': reference})
        if secure:
            r.pop('auth_data', None)

        return r


    INDEX_SCHEMA    = xys.load("""
    offset?:        !~~digit
    limit?:         !~~uint
    filter?:        !!any
    sort_by?:       !~~isIn(vpn_client.sort_by)
    sort_order?:    !~~isIn(vpn_client.order_by)
    """)

    INDEX_DEFAULTS  = {'offset':        0,
                       'limit':         10,
                       'filter':        {},
                       'sort_by':       'vpn_client.id',
                       'sort_order':    'asc'}

    def index(self, request):
        options = request.query_params()
        if not xys.validate(options, self.INDEX_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        for key, value in iteritems(self.INDEX_DEFAULTS):
            options.setdefault(key, value)

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        r       = {'results':   [],
                   'total':     0}

        try:
            cnt = self.objects['vpn_client'].get_count()
            res = self.objects['vpn_client'].get_all_server_user(to_dict = True,
                                                                 order   = ((options['sort_by'], options['sort_order']),),
                                                                 limit   = int(options['limit']),
                                                                 offset  = int(options['offset']))
            if res:
                for row in res:
                    row['vpn_client.ip_addr'] = '.'.join(row['vpn_server.cidr'].split('.')[0:3] + [row['vpn_client.last_octet']])
                    row.pop('vpn_client.auth_data', None)

                r['results']    = res
                r['total']      = cnt

            return r
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


    FREE_LAST_OCTETS_SCHEMA  = xys.load("""
    server_id: !~~digit
    client_id: !~~digit
    """)

    def free_last_octets(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self.FREE_LAST_OCTETS_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)
        try:
            server      = self.objects['vpn_server'].get(**{'id': options['server_id']})
            if not server:
                raise HttpReqErrJson(404, "unable to find server. (id: %r)" % options['server_id'])

            server_ip   = '.'.join(server['cidr'].split('.')[0:3])
            last_octets = self.objects['vpn_client'].get_free_last_octets(server['id'], options['client_id'])

            if not last_octets:
                raise HttpReqErrJson(507, "no more free ip address available")

            return [{'last_octet':  last_octet,
                     'addr_ip':     "%s.%s" % (server_ip, last_octet)} for last_octet in last_octets]
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


    FORM_SCHEMA  = xys.load("""
    id?:    !~~digit
    action: !~~enum(create,update)
    """)

    def form(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqError(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self.FORM_SCHEMA):
            raise HttpReqError(415, "invalid arguments for command")

        servers = self.objects['vpn_server'].get_all_user(to_dict   = True,
                                                          order     = 'vpn_server.id',
                                                          with_user = True)

        if options['action'] == 'create':
            return self.render('form.html',
                               action       = 'create',
                               xid          = '',
                               auth_methods = FdVpnObjectVPNClient.get_auth_method_list(),
                               vpn_client   = {},
                               servers      = servers or [])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqError(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            xobject             = self._get(options['id'])
            if not xobject:
                raise HttpReqError(404, "client not found. (id: %r)" % options['id'])

            server              = self.objects['vpn_server'].get(**{'id': xobject['server_id']})
            if not server:
                raise HttpReqError(404, "server not found. (id: %r)" % xobject['server_id'])

            xobject['ip_addr']  = '.'.join(server['cidr'].split('.')[0:3] + [xobject['last_octet']])

            return self.render('form.html',
                               action       = 'update',
                               xid          = options['id'],
                               auth_methods = FdVpnObjectVPNClient.get_auth_method_list(),
                               vpn_client   = xobject,
                               servers      = servers or [])
        except HttpReqError:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqError(503, repr(e))
        finally:
            self.LOCK.release()


    GET_SCHEMA  = xys.load("""
    id: !~~digit
    """)

    def get(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self.GET_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            xobject             = self._get(options['id'], secure = False)
            if not xobject:
                raise HttpReqErrJson(404, "client not found. (id: %r)" % options['id'])

            server              = self.objects['vpn_server'].get(**{'id': xobject['server_id']})
            if not server:
                raise HttpReqErrJson(404, "server not found. (id: %r)" % xobject['server_id'])

            xobject['username'] = None
            xobject['user_id']  = None
            xobject['auth_qrcode'] = None

            if server['user_id']:
                user = self.objects['vpn_user'].get(**{'id': server['user_id']})
                if user:
                    xobject['username'] = user['username']
                    xobject['user_id']  = user['id']

            xobject['ip_addr']  = '.'.join(server['cidr'].split('.')[0:3] + [xobject['last_octet']])

            if xobject['auth_method'] == 'otp':
                secret_url = pyotp.totp.TOTP(xobject['auth_data']).provisioning_uri(
                    name=xobject['reference'],
                    issuer_name='vpn.supercmd.io')

                buffered = BytesIO()
                img = qrcode.make(secret_url)
                img.save(buffered, format = 'PNG')
                xobject['auth_qrcode'] = base64.b64encode(buffered.getvalue())

            xobject.pop('auth_data', None)

            return xobject
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def _generate_installer(self, server, client, arch, files):
        build_dir = None
        cwd       = os.getcwd()

        try:
            build_dir   = tempfile.mkdtemp()
            os.chdir(build_dir)

            confvpn_dir = os.path.join(build_dir, 'OpenVPN', 'config')

            config_dir  = os.path.join(confvpn_dir, server['reference'])
            os.makedirs(config_dir)

            config_file = os.path.join(confvpn_dir, "%s.ovpn" % client['reference'])
            config_inline_file = os.path.join(confvpn_dir, "%s-inline.ovpn" % client['reference'])

            """
            helpers.file_w_tmp(self.plugin_openvpn.get_client_config(client['reference'],
                                                                     server).getvalue(),
                               config_file)
            """

            with open(config_file, 'w+') as f:
                f.write(self.plugin_openvpn.get_client_config(client['reference'],
                                                              client['auth_method'],
                                                              server).getvalue())

            with open(config_inline_file, 'w+') as f:
                f.write(self.plugin_openvpn.get_client_config_inline(client['reference'],
                                                                     client['auth_method'],
                                                                     server,
                                                                     files).getvalue())

            for xfile in itervalues(files):
                shutil.copy2(xfile, config_dir)

            if arch == 'ovpn':
                os.chdir(confvpn_dir)
                outfile = os.path.join(build_dir, self.config['vpn']['client']['zip_outfile'])
                zipf = zipfile.ZipFile(outfile, 'w', zipfile.ZIP_DEFLATED)
                for wroot, wdirs, wfiles in os.walk('.'):
                    for wfile in wfiles:
                        zipf.write(os.path.join(wroot, wfile))
                zipf.close()

                with open(outfile, 'rb') as f:
                    outdata = f.read()

                return HttpResponse(data    = outdata,
                                    headers = {'Content-type': 'application/zip',
                                               'Content-Disposition': 'inline; filename="%s-%s-%s.zip"'
                                                                       % (client['reference'],
                                                                          arch,
                                                                          self.config['vpn']['client']['zip_outfile'])})

            exe_file    = self.config['vpn']['client']['openvpn_exe'][arch]

            shutil.copyfile(exe_file, os.path.join(build_dir, OPENVPN_EXE_NAME))

            cmd         = subprocess.Popen(shlex.split(self.config['vpn']['client']['nsi_cmd']
                                                       % {'build_dir':        build_dir,
                                                          'client_reference': client['reference'],
                                                          'server_reference': server['reference'],
                                                          'cidr':             server['cidr']}),
                                           stderr       = subprocess.PIPE,
                                           close_fds    = True)
            (out, err)  = cmd.communicate()

            if cmd.returncode:
                raise RuntimeError("unable to generate installer. (build_dir: %r, err: %r, rc: %r)"
                                   % (build_dir, err, cmd.returncode))

            outfile     = os.path.join(build_dir, self.config['vpn']['client']['nsi_outfile'])

            with open(outfile, 'rb') as f:
                outdata = f.read()

            return HttpResponse(data    = outdata,
                                headers = {'Content-type': 'application/octet-stream',
                                           'Content-Disposition': 'inline; filename="%s-%s-%s"'
                                                                   % (client['reference'],
                                                                      arch,
                                                                      self.config['vpn']['client']['nsi_outfile'])})
        finally:
            if os.path.isdir(cwd):
                os.chdir(cwd)
            if build_dir:
                shutil.rmtree(build_dir)


    DOWNLOAD_SCHEMA  = xys.load("""
    id:     !~~digit
    arch:   !~~enum(x32,x64,xp_x32,xp_x64,ovpn)
    """)

    def download(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options         = request.query_params()
        if not xys.validate(options, self.DOWNLOAD_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        client = self._get(options['id'])
        if not client:
            raise HttpReqErrJson(404, "unable to find client. (id: %r)" % options['id'])

        server          = self.objects['vpn_server'].get(to_dict = True,
                                                         **{'id': client['server_id']})
        if not server:
            raise HttpReqErrJson(404, "unable to find server. (id: %r)" % client['server_id'])

        if not server['user_id']:
            raise HttpReqErrJson(415, "invalid server. (id: %r)" % server['id'])

        obj             = {}
        obj.update(dict(["vpn_client.%s" % x, y] for x, y in iteritems(client)))
        obj.update(dict(["vpn_server.%s" % x, y] for x, y in iteritems(server)))

        ca_crt_path     = os.path.join(self.ca_path, "%s.crt" % self.ca_name)
        ca_pkey_path    = os.path.join(self.ca_path, "%s.key" % self.ca_name)
        ca_srl_path     = os.path.join(self.ca_path, "%s.srl" % self.ca_name)

        if not os.access(ca_crt_path, (os.R_OK | os.F_OK)):
            raise HttpReqErrJson(400, "unable to find the CA certificate: %r" % ca_crt_path)

        if not os.access(ca_pkey_path, (os.R_OK | os.F_OK)):
            raise HttpReqErrJson(400, "unable to find the PKey certificate: %r" % ca_pkey_path)

        if not os.access(ca_srl_path, (os.R_OK | os.F_OK)):
            raise HttpReqErrJson(400, "unable to find the serial number: %r" % ca_srl_path)

        attributes                  = self.config['vpn']['client']['attributes'].copy()
        attributes['commonName']    = self.common_name % obj

        client_pkey_path            = os.path.join(self.client_path, "%s.key" % attributes['commonName'])
        client_crt_path             = os.path.join(self.client_path, "%s.crt" % attributes['commonName'])

        config                      = self.plugin_openvpn.get_server_config(server['reference'])[1]

        files                       = {'ca':        ca_crt_path,
                                       'pkey':      client_pkey_path,
                                       'crt':       client_crt_path,
                                       'ta_key':    config.get('general', 'tls-auth').split(' ', 1)[0]}

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        error           = False
        client_csr_path = None

        try:
            if os.access(client_pkey_path, (os.R_OK | os.F_OK)) \
               and os.access(client_crt_path, (os.R_OK | os.F_OK)):
                with open(client_pkey_path, 'r') as f:
                    client_pkey = crypto.load_privatekey(crypto.FILETYPE_PEM,
                                                         ''.join(f.readlines()))

                with open(client_crt_path, 'r') as f:
                    client_crt  = crypto.load_certificate(crypto.FILETYPE_PEM,
                                                          ''.join(f.readlines()))

                if not client_crt.has_expired():
                    crypto.dump_privatekey(crypto.FILETYPE_PEM, client_pkey)
                    crypto.dump_certificate(crypto.FILETYPE_PEM, client_crt)

                    return self._generate_installer(server,
                                                    client,
                                                    options['arch'],
                                                    files)

            client_csr_path = os.path.join(self.client_path, "%s.csr" % attributes['commonName'])

            with open(ca_crt_path, 'r') as f:
                ca_crt  = crypto.load_certificate(crypto.FILETYPE_PEM,
                                                  ''.join(f.readlines()))

            with open(ca_pkey_path, 'r') as f:
                ca_pkey = crypto.load_privatekey(crypto.FILETYPE_PEM,
                                                 ''.join(f.readlines()))

            is_hex = False

            with open(ca_srl_path, 'r') as f:
                x = f.readline().strip()
                is_hex = not x.isdigit()
                if is_hex:
                    ca_srl = int(x, 16) + 1
                else:
                    ca_srl = int(x) + 1

            gen_cert        = GenCert(self.config['vpn']['client']['options']['bits'],
                                      self.config['vpn']['client']['options']['crypto'],
                                      self.config['vpn']['client']['options']['digest'],
                                      notafter_days = self.config['vpn']['client']['options']['days'])

            pkey            = gen_cert.make_privatekey(client_pkey_path)
            csr             = gen_cert.make_certreq(pkey,
                                                    attributes,
                                                    client_csr_path,
                                                    keyUsage = ['digitalSignature', 'keyAgreement'],
                                                    extendedKeyUsage = ['clientAuth'])

            crt             = gen_cert.make_certificate(csr,
                                                        ca_crt,
                                                        ca_pkey,
                                                        ca_srl,
                                                        client_crt_path)

            crypto.dump_privatekey(crypto.FILETYPE_PEM, pkey)
            crypto.dump_certificate(crypto.FILETYPE_PEM, crt)

            if is_hex:
                srl = format(crt.get_serial_number(), 'x').upper()
            else:
                srl = crt.get_serial_number()

            helpers.file_w_tmp("%s" % srl, ca_srl_path, mode = 'w+')

            return self._generate_installer(server,
                                            client,
                                            options['arch'],
                                            files)
        except HttpReqErrJson:
            error = True
            raise
        except Exception as e:
            error = True
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            if error:
                if os.path.exists(client_crt_path):
                    os.unlink(client_crt_path)
                if os.path.exists(client_pkey_path):
                    os.unlink(client_pkey_path)
                if os.path.exists(client_csr_path):
                    os.unlink(client_csr_path)
            self.LOCK.release()


    CREATE_SCHEMA   = xys.load("""
    server_id:         !~~digit
    last_octet:        !~~digit
    auth_method?:      !~~isIn(vpn_client.auth_method)
    auth_data?-:       !!str
    comment*[0,1000]-: !!str
    """)

    def create(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        args = request.payload_params()
        if not xys.validate(args, self.CREATE_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            server              = self.objects['vpn_server'].get(**{'id': args['server_id']})
            if not server:
                raise HttpReqErrJson(404, "unable to find server. (id: %r)" % args['server_id'])

            if not server['user_id']:
                raise HttpReqErrJson(415, "invalid server. (id: %r)" % server['id'])

            last_octets         = self.objects['vpn_client'].get_free_last_octets(server['id'])
            if not last_octets:
                raise HttpReqErrJson(507, "no more free ip address available")

            if int(args['last_octet']) not in last_octets:
                raise HttpReqErrJson(415, "invalid address ip. (last_octet: %r)" % args['last_octet'])

            data                = dict(args)
            data['id']          = None
            data['reference']   = str(uuid.uuid4())[0:8]
            data['enabled']     = True

            if data['auth_method'] == 'otp':
                data['auth_data'] = pyotp.random_base32()
            elif data['auth_method'] == 'userpass':
                try:
                    helpers.password_check(data['auth_data'])
                except Exception as e:
                    raise HttpReqErrJson(415, repr(e))
                data['auth_data'] = bcrypt.hashpw(ensure_str(data['auth_data']),
                                                  bcrypt.gensalt())
            else:
                data['auth_data'] = None

            xobject             = self._get(
                                    self.objects['vpn_client'].create(
                                        data.keys(),
                                        data.values()))

            if not int(server['enabled']) or not server['user_id']:
                self.plugin_iptables.add_drop_src('SERVER', server['cidr'])
                self.plugin_iptables.add_drop_dst('SERVER', server['cidr'])
            else:
                self.plugin_iptables.delete_drop_src('SERVER', server['cidr'])
                self.plugin_iptables.delete_drop_dst('SERVER', server['cidr'])

            self.plugin_openvpn.create_server_config(
                self.plugin_openvpn.build_server_config(server))

            self.plugin_openvpn.create_ccd_config(server['reference'],
                                                  server['cidr'],
                                                  xobject)

            self.plugin_openvpn.cmd_call('enable', server['reference'])
            #self.plugin_openvpn.cmd_call('restart', server['reference'])
            return xobject
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def _update(self, xid, data):
        return self.objects['vpn_client'].update(data.keys(),
                                                 data.values(),
                                                 **{'id': xid})


    UPDATE_SCHEMA   = xys.load("""
    id:                !~~digit
    user_id?:          !~~digit
    server_id:         !~~digit
    last_octet:        !~~digit
    auth_method?:      !~~isIn(vpn_client.auth_method)
    auth_data?-:       !!str
    comment*[0,1000]-: !!str
    username?:         !!str
    ip_addr?:          !!str
    """)

    def update(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        args = request.payload_params()
        if not xys.validate(args, self.UPDATE_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            server      = self.objects['vpn_server'].get(**{'id': args['server_id']})
            if not server:
                raise HttpReqErrJson(404, "unable to find server. (id: %r)" % args['server_id'])

            if not server['user_id']:
                raise HttpReqErrJson(415, "invalid server. (id: %r)" % server['id'])

            xid         = args.pop('id')
            xdict       = {'server_id':     args['server_id'],
                           'last_octet':    args['last_octet'],
                           'auth_method':   args['auth_method'],
                           'comment':       args['comment']}

            xobject     = self._get(xid, secure = False)
            if not xobject:
                raise HttpReqErrJson(404, "unable to find client. (id: %r)" % xid)

            if xdict['auth_method'] == 'otp':
                if xobject['auth_method'] != xdict['auth_method']:
                    xdict['auth_data'] = pyotp.random_base32()
                elif not xobject['auth_data']:
                    xdict['auth_data'] = pyotp.random_base32()
            elif xdict['auth_method'] == 'userpass':
                if xobject['auth_method'] != xdict['auth_method'] \
                   and not args.get('auth_data'):
                    raise HttpReqErrJson(415, "missing password")

                if args.get('auth_data'):
                    try:
                        helpers.password_check(args['auth_data'])
                    except Exception as e:
                        raise HttpReqErrJson(415, repr(e))
                    xdict['auth_data'] = bcrypt.hashpw(ensure_str(args['auth_data']),
                                                       bcrypt.gensalt())
            else:
                xdict['auth_data'] = None

            last_octets = self.objects['vpn_client'].get_free_last_octets(server['id'], xobject['id'])
            if not last_octets:
                raise HttpReqErrJson(507, "no more free ip address available")

            if int(args['last_octet']) not in last_octets \
               and int(args['last_octet']) != int(xobject['last_octet']):
                raise HttpReqErrJson(415, "invalid address ip. (last_octet: %r)" % args['last_octet'])

            if xobject['server_id']:
                server_old      = self.objects['vpn_server'].get(**{'id': xobject['server_id']})
                if server_old:
                    self.plugin_openvpn.remove_ccd_config(server_old['reference'],
                                                          xobject)

            if not self._update(xid, xdict):
                raise HttpReqErrJson(500, "unable to update client. (id: %r)" % xid)

            xobject     = self._get(xid)

            if not int(server['enabled']) or not server['user_id']:
                self.plugin_iptables.add_drop_src('SERVER', server['cidr'])
                self.plugin_iptables.add_drop_dst('SERVER', server['cidr'])
            else:
                self.plugin_iptables.delete_drop_src('SERVER', server['cidr'])
                self.plugin_iptables.delete_drop_dst('SERVER', server['cidr'])

            self.plugin_openvpn.create_server_config(
                self.plugin_openvpn.build_server_config(server))

            self.plugin_openvpn.cmd_call('enable', server['reference'])
            #self.plugin_openvpn.cmd_call('restart', server['reference'])

            self.plugin_openvpn.create_ccd_config(server['reference'],
                                                  server['cidr'],
                                                  xobject)
            return xobject
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


    _TOGGLE_ENABLED_SCHEMA  = xys.load("""
    id: !~~digit
    """)

    def _toggle_enabled(self, request, enabled):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self._TOGGLE_ENABLED_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            xid = options.pop('id')

            self._update(xid, {'enabled': helpers.boolize(enabled)})

            xobject = self._get(xid)
            if not xobject:
                raise HttpReqErrJson(404, "unable to find client. (id: %r)" % xid)

            server  = self.objects['vpn_server'].get(**{'id': xobject['server_id']})
            if not server:
                raise HttpReqErrJson(404, "unable to find server. (id: %r)" % xobject['server_id'])

            addr_ip = '.'.join(server['cidr'].split('.')[0:3] + [xobject['last_octet']])

            if not int(xobject['enabled']):
                self.plugin_iptables.add_drop_src('CLIENT', addr_ip)
                self.plugin_iptables.add_drop_dst('CLIENT', addr_ip)
                self.plugin_openvpn.disable_ccd_config(server['reference'], xobject)
            else:
                self.plugin_iptables.delete_drop_src('CLIENT', addr_ip)
                self.plugin_iptables.delete_drop_dst('CLIENT', addr_ip)
                self.plugin_openvpn.enable_ccd_config(server['reference'], xobject)

            return xobject
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def enable(self, request):
        return self._toggle_enabled(request, True)

    def disable(self, request):
        return self._toggle_enabled(request, False)


    DELETE_SCHEMA  = xys.load("""
    id: !~~digit
    """)

    def delete(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self.DELETE_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            xid = options.pop('id')

            xobject = self._get(xid)
            if not xobject:
                raise HttpReqErrJson(404, "unable to find client. (id: %r)" % xid)

            server  = self.objects['vpn_server'].get(**{'id': xobject['server_id']})
            if not server:
                raise HttpReqErrJson(404, "unable to find server. (id: %r)" % xobject['server_id'])

            self.plugin_openvpn.remove_ccd_config(server['reference'], xobject)
            self.objects['vpn_client'].delete(**{'id': xid})
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


    _TOGGLE_CONNECTED_SCHEMA  = xys.load("""
    reference: !~~regex(vpn_client.reference)
    """)

    def _toggle_connected(self, request, connected):
        options = request.query_params()
        if not xys.validate(options, self._TOGGLE_CONNECTED_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            reference = options.pop('reference')

            xdict = {'connected': helpers.boolize(connected)}

            if connected:
                xdict['connected_at'] = time.strftime('%Y-%m-%d %H:%M:%S')
            else:
                xdict['disconnected_at'] = time.strftime('%Y-%m-%d %H:%M:%S')

            xobject = self._get_by_reference(reference)
            if not xobject:
                raise HttpReqErrJson(404, "unable to find client. (reference: %r)" % reference)

            if not xobject['enabled']:
                raise HttpReqErrJson(404, "client disabled. (reference: %r)" % reference)

            self._update(xobject['id'], xdict)

            server  = self.objects['vpn_server'].get(**{'id': xobject['server_id']})
            if not server:
                raise HttpReqErrJson(404, "unable to find server. (id: %r)" % xobject['server_id'])

            if not server['enabled']:
                raise HttpReqErrJson(404, "server disabled. (reference: %r)" % server['reference'])

            return True
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def connect(self, request):
        return self._toggle_connected(request, True)

    def disconnect(self, request):
        return self._toggle_connected(request, False)


    AUTHENTICATE_QSCHEMA  = xys.load("""
    reference: !~~regex(vpn_client.reference)
    """)

    AUTHENTICATE_PSCHEMA  = xys.load("""
    password-: !!str
    """)

    def authenticate(self, request):
        options = request.query_params()
        if not xys.validate(options, self.AUTHENTICATE_QSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        args = request.payload_params()
        if not xys.validate(args, self.AUTHENTICATE_PSCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            reference = options.pop('reference')

            xobject = self._get_by_reference(reference, secure = False)
            if not xobject:
                raise HttpReqErrJson(404, "unable to find client. (reference: %r)" % reference)

            if not xobject['enabled']:
                raise HttpReqErrJson(404, "client disabled. (reference: %r)" % reference)

            server  = self.objects['vpn_server'].get(**{'id': xobject['server_id']})
            if not server:
                raise HttpReqErrJson(404, "unable to find server. (id: %r)" % xobject['server_id'])

            if not server['enabled']:
                raise HttpReqErrJson(404, "server disabled. (reference: %r)" % server['reference'])

            if xobject['auth_method'] == 'userpass':
                return bcrypt.checkpw(ensure_str(args['password']),
                                      ensure_str(xobject['auth_data']))

            if xobject['auth_method'] == 'otp':
                return pyotp.totp.TOTP(xobject['auth_data']).verify(ensure_str(args['password']))

            return False
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


    VALIDATE_SCHEMA = xys.load("""
    server_id:         !~~digit
    last_octet:        !~~digit
    auth_method?:      !~~isIn(vpn_client.auth_method)
    comment?[0,1000]-: !!str
    """)

    def validate(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            for key, value in iteritems(options):
                if not self.VALIDATE_SCHEMA.has_key(key):
                    raise HttpReqErrJson(404, "unknown argument for command. (key: %r)" % key)

                if not xys.validate({key: value}, {key: self.VALIDATE_SCHEMA[key]}):
                    raise HttpReqErrJson(415, "invalid argument for command. (key: %r)" % key)

            return True
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


if __name__ != "__main__":
    def _start():
        MODULES.register(VPNClientModule())
    _start()
