# -*- coding: utf-8 -*-
"""vpn_server module"""

import logging
import re

from six import iteritems

from fd_vpn.objects.vpn_client import FdVpnObjectVPNClient
from fd_vpn.objects.vpn_server import FdVpnObjectVPNServer
from fd_vpn.objects.vpn_user import FdVpnObjectVPNUser

from dwho.classes.modules import DWhoModuleSQLBase, DWhoModuleWebBase, MODULES
from dwho.classes.plugins import PLUGINS
from sonicprobe import helpers
from sonicprobe.libs import xys
from sonicprobe.libs.moresynchro import RWLock
from httpdis.ext.httpdis_json import HttpReqError, HttpReqErrJson


LOG = logging.getLogger('fd-vpn.modules.server')

xys.add_list('vpn_server.sort_by', FdVpnObjectVPNServer.get_columns() + ['vpn_user.username'])
xys.add_list('vpn_server.order_by', ('asc', 'desc'))
xys.add_regex('vpn_server.reference', re.compile(r'^[a-z0-9]{8}$').match)


class VPNServerModule(DWhoModuleSQLBase, DWhoModuleWebBase):
    MODULE_NAME = 'vpn_server'

    LOCK        = RWLock()

    def __init__(self):
        self.objects    = {'vpn_client': FdVpnObjectVPNClient(),
                           'vpn_server': FdVpnObjectVPNServer(),
                           'vpn_user':   FdVpnObjectVPNUser()}

        DWhoModuleSQLBase.__init__(self)
        DWhoModuleWebBase.__init__(self)

    def init(self, config):
        DWhoModuleSQLBase.init(self, config)
        DWhoModuleWebBase.init(self, config)

        self.objects['vpn_client'].init(config)
        self.objects['vpn_server'].init(config)
        self.objects['vpn_user'].init(config)

    def safe_init(self, options):
        self.lock_timeout       = self.config['general']['lock_timeout']

        self.plugin_iptables    = PLUGINS['iptables']
        self.plugin_openvpn     = PLUGINS['openvpn']
        self.plugin_credentials = PLUGINS['credentials']

    def at_start(self, options):
        servers = self.objects['vpn_server'].get_all()

        if not servers:
            return

        for server in servers:
            if not int(server['enabled']) or not server['user_id']:
                self.plugin_iptables.add_drop_sport('SERVER', server['port_number'])
                self.plugin_iptables.add_drop_dport('SERVER', server['port_number'])
            else:
                self.plugin_iptables.delete_drop_sport('SERVER', server['port_number'])
                self.plugin_iptables.delete_drop_dport('SERVER', server['port_number'])

            """"
            clients = self.get_server_clients(server['id'])
            if not clients:
                continue

            if not int(server['enabled']) or not server['user_id']:
                self.plugin_openvpn.disable_ccd_config
            """

    def get_server_clients(self, server_id, available = True, selected = True):
        r       = {'selected':  None,
                   'lists':     {'available':   [],
                                 'selected':    []}}

        if available:
            clients = self.objects['vpn_client'].get_all(to_dict    = True,
                                                         order      = 'last_octet',
                                                         **{'server_id': None})
            if clients:
                for item in clients:
                    r['lists']['available'].append({'id':         item['id'],
                                                    'last_octet': item['last_octet'],
                                                    'reference':  item['reference'],
                                                    'enabled':    item['enabled'],
                                                    'ip_addr':    None})

        if not server_id or not selected:
            return r

        clients = self.objects['vpn_client'].get_all(to_dict    = True,
                                                     order      = 'last_octet',
                                                     **{'server_id': server_id})

        if not clients:
            return r

        server  = self.objects['vpn_server'].get(**{'id': server_id})

        for item in clients:
            r['lists']['selected'].append({'id':         item['id'],
                                           'last_octet': item['last_octet'],
                                           'reference':  item['reference'],
                                           'enabled':    item['enabled'],
                                           'ip_addr':    '.'.join(server['cidr'].split('.')[0:3]
                                                            + [str(item['last_octet'])])})

        return r


    INDEX_SCHEMA    = xys.load("""
    offset?:        !~~digit
    limit?:         !~~uint
    filter?:        !!any
    sort_by?:       !~~isIn(vpn_server.sort_by)
    sort_order?:    !~~isIn(vpn_server.order_by)
    """)

    INDEX_DEFAULTS  = {'offset':     0,
                       'limit':      10,
                       'filter':     {},
                       'sort_by':    'vpn_server.id',
                       'sort_order': 'asc'}

    def index(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self.INDEX_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        for key, value in iteritems(self.INDEX_DEFAULTS):
            options.setdefault(key, value)

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        r       = {'results': [],
                   'total':   0}

        try:
            cnt = self.objects['vpn_server'].get_count()
            res = self.objects['vpn_server'].get_all_user(to_dict = True,
                                                          order   = ((options['sort_by'], options['sort_order']),),
                                                          limit   = int(options['limit']),
                                                          offset  = int(options['offset']))
            if res:
                r['results'] = res
                r['total']   = cnt

            return r
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def _get(self, xid):
        return self.objects['vpn_server'].get(to_dict = True,
                                              **{'id': xid})


    FORM_SCHEMA  = xys.load("""
    id:     !~~digit
    action: !~~enum(update)
    """)

    def form(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqError(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self.FORM_SCHEMA):
            raise HttpReqError(415, "invalid arguments for command")

        users   = self.objects['vpn_user'].get_all(to_dict  = True,
                                                   order    = 'username')

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqError(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            xobject = self._get(options['id'])
            if not xobject:
                raise HttpReqError(404, "server not found. (id: %r)" % options['id'])

            return self.render('form.html',
                               action   = 'update',
                               xid      = options['id'],
                               server   = xobject,
                               users    = users or [])
        except HttpReqError:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqError(503, repr(e))
        finally:
            self.LOCK.release()


    GET_SCHEMA  = xys.load("""
    id: !~~digit
    """)

    def get(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self.GET_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            xobject             = self._get(options['id'])
            if not xobject:
                raise HttpReqErrJson(404, "server not found. (id: %r)" % options['id'])

            xobject['clients']  = self.get_server_clients(options['id'])
            xobject['username'] = None

            if xobject['user_id']:
                user = self.objects['vpn_user'].get(**{'id': xobject['user_id']})
                if user:
                    xobject['username'] = user['username']

            return xobject
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def _update(self, xid, data):
        return self.objects['vpn_server'].update(data.keys(),
                                                 data.values(),
                                                 **{'id': xid})


    UPDATE_SCHEMA   = xys.load("""
    id:             !~~digit
    user_id:        !~~digit
    """)

    def update(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        args = request.payload_params()
        if not xys.validate(args, self.UPDATE_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            if not self.objects['vpn_user'].exists(**{'id': args['user_id']}):
                raise HttpReqErrJson(404, "unable to find user. (id: %r)" % args['user_id'])

            xid             = args.pop('id')
            xobject         = self._get(xid)
            if not xobject:
                raise HttpReqErrJson(404, "unable to find server. (id: %r)" % xid)

            if not self._update(xid, {'user_id': args['user_id']}):
                raise HttpReqErrJson(500, "unable to update server. (id: %r)" % xid)

            if not int(xobject['enabled']):
                self.plugin_iptables.add_drop_sport('SERVER', xobject['port_number'])
                self.plugin_iptables.add_drop_dport('SERVER', xobject['port_number'])
            else:
                self.plugin_iptables.delete_drop_sport('SERVER', xobject['port_number'])
                self.plugin_iptables.delete_drop_dport('SERVER', xobject['port_number'])

            self.plugin_openvpn.create_server_config(
                self.plugin_openvpn.build_server_config(xobject))

            self.plugin_openvpn.cmd_call('enable', xobject['reference'])
            self.plugin_openvpn.cmd_call('restart', xobject['reference'])

            r               = self._get(xobject['id'])
            r['clients']    = self.get_server_clients(r['id'])

            return r
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


    _TOGGLE_ENABLED_SCHEMA  = xys.load("""
    id: !~~digit
    """)

    def _toggle_enabled(self, request, enabled):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self._TOGGLE_ENABLED_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            xid     = options.pop('id')

            self._update(xid, {'enabled': helpers.boolize(enabled)})

            xobject = self._get(xid)
            if not xobject:
                raise HttpReqErrJson(404, "unable to find server. (id: %r)" % xid)

            if not int(xobject['enabled']):
                self.plugin_iptables.add_drop_sport('SERVER', xobject['port_number'])
                self.plugin_iptables.add_drop_dport('SERVER', xobject['port_number'])
            else:
                self.plugin_iptables.delete_drop_sport('SERVER', xobject['port_number'])
                self.plugin_iptables.delete_drop_dport('SERVER', xobject['port_number'])

            return xobject
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def enable(self, request):
        return self._toggle_enabled(request, True)

    def disable(self, request):
        return self._toggle_enabled(request, False)


    VALIDATE_SCHEMA = xys.load("""
    user_id:    !~~digit
    """)

    def validate(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            for key, value in options.iteritems():
                if not self.VALIDATE_SCHEMA.has_key(key):
                    raise HttpReqErrJson(404, "unknown argument for command. (key: %r)" % key)
                if not xys.validate({key: value}, {key: self.VALIDATE_SCHEMA[key]}):
                    raise HttpReqErrJson(415, "invalid argument for command. (key: %r)" % key)

            return True
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


if __name__ != "__main__":
    def _start():
        MODULES.register(VPNServerModule())
    _start()
