# -*- coding: utf-8 -*-
"""vpn_user module"""

import logging
import re
import time

from six import iteritems

from fd_vpn.objects.vpn_server import FdVpnObjectVPNServer
from fd_vpn.objects.vpn_user import FdVpnObjectVPNUser

from dwho.classes.modules import DWhoModuleSQLBase, DWhoModuleWebBase, MODULES
from dwho.classes.plugins import PLUGINS
from sonicprobe import helpers
from sonicprobe.libs import xys
from sonicprobe.libs.moresynchro import RWLock
from httpdis.ext.httpdis_json import HttpReqError, HttpReqErrJson


LOG          = logging.getLogger('fd-vpn.modules.user')

xys.add_list('vpn_user.sort_by', FdVpnObjectVPNUser.COLUMNS)
xys.add_list('vpn_user.order_by', ('asc', 'desc'))
xys.add_regex('vpn_user.username', re.compile(r'^[a-z][a-zA-Z0-9\-\.\_@]{3,255}$').match)


class VPNUserModule(DWhoModuleSQLBase, DWhoModuleWebBase):
    MODULE_NAME = 'vpn_user'

    LOCK        = RWLock()

    def __init__(self):
        self.objects    = {'vpn_server': FdVpnObjectVPNServer(),
                           'vpn_user':   FdVpnObjectVPNUser()}

        DWhoModuleSQLBase.__init__(self)
        DWhoModuleWebBase.__init__(self)

    def init(self, config):
        DWhoModuleSQLBase.init(self, config)
        DWhoModuleWebBase.init(self, config)

        self.objects['vpn_server'].init(config)
        self.objects['vpn_user'].init(config)

    def safe_init(self, options):
        self.lock_timeout       = self.config['general']['lock_timeout']

        self.plugin_iptables    = PLUGINS['iptables']
        self.plugin_openvpn     = PLUGINS['openvpn']
        self.plugin_credentials = PLUGINS['credentials']

    def at_start(self, options):
        servers = self.objects['vpn_user'].get_servers()

        if not servers:
            return

        for server in servers:
            if not int(server['vpn_user.enabled']):
                self.plugin_iptables.add_drop_sport('USER', server['vpn_server.port_number'])
                self.plugin_iptables.add_drop_dport('USER', server['vpn_server.port_number'])
            else:
                self.plugin_iptables.delete_drop_sport('USER', server['vpn_server.port_number'])
                self.plugin_iptables.delete_drop_dport('USER', server['vpn_server.port_number'])

    def get_user_servers(self, user_id, available = True, selected = True):
        r       = {'selected':  None,
                   'lists':     {'available':   [],
                                 'selected':    []}}

        if available:
            servers = self.objects['vpn_server'].get_all(to_dict    = True,
                                                         order      = (('port_number', 'asc'),
                                                                       ('cidr', 'asc')),
                                                         **{'user_id': None})
            if servers:
                for item in servers:
                    r['lists']['available'].append({'id':           item['id'],
                                                    'cidr':         item['cidr'],
                                                    'port_number':  item['port_number'],
                                                    'enabled':      item['enabled'],
                                                    'reference':    item['reference']})

        if not user_id or not selected:
            return r

        servers = self.objects['vpn_server'].get_all(to_dict    = True,
                                                     order      = (('port_number', 'asc'),
                                                                   ('cidr', 'asc')),
                                                     **{'user_id': user_id})

        if not servers:
            return r

        for item in servers:
            r['lists']['selected'].append({'id':            item['id'],
                                           'cidr':          item['cidr'],
                                           'port_number':   item['port_number'],
                                           'enabled':       item['enabled'],
                                           'reference':     item['reference']})

        return r


    SERVERS_SCHEMA  = xys.load("""
    id: !~~digit
    """)

    def servers(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self.SERVERS_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        options['id'] = int(options['id'])

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            return self.get_user_servers(options['id'])
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


    INDEX_SCHEMA    = xys.load("""
    offset?:        !~~digit
    limit?:         !~~uint
    filter?:        !!any
    sort_by?:       !~~isIn(vpn_user.sort_by)
    sort_order?:    !~~isIn(vpn_user.order_by)
    """)

    INDEX_DEFAULTS  = {'offset':        0,
                       'limit':         10,
                       'filter':        {},
                       'sort_by':       'id',
                       'sort_order':    'asc'}

    def index(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self.INDEX_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        for key, value in iteritems(self.INDEX_DEFAULTS):
            options.setdefault(key, value)

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        r       = {'results':   [],
                   'total':     0}

        try:
            cnt = self.objects['vpn_user'].get_count()
            res = self.objects['vpn_user'].get_all(to_dict  = True,
                                                   order    = ((options['sort_by'], options['sort_order']),),
                                                   limit    = int(options['limit']),
                                                   offset   = int(options['offset']))
            if res:
                r['results']    = res
                r['total']      = cnt

            return r
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def _get(self, xid):
        return self.objects['vpn_user'].get(to_dict = True,
                                            **{'id': xid})


    FORM_SCHEMA  = xys.load("""
    id?:    !~~digit
    action: !~~enum(create,update)
    """)

    def form(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqError(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self.FORM_SCHEMA):
            raise HttpReqError(415, "invalid arguments for command")

        if options['action'] == 'create':
            return self.render('form.html',
                               action   = 'create',
                               xid      = '')

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqError(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            if not self.objects['vpn_user'].exists(**{'id': options['id']}):
                raise HttpReqError(404, "object not found. (id: %r)" % options['id'])

            return self.render('form.html',
                               action   = 'update',
                               xid      = options['id'])
        except HttpReqError:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqError(503, str(e))
        finally:
            self.LOCK.release()


    GET_SCHEMA  = xys.load("""
    id: !~~digit
    """)

    def get(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self.GET_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            xobject             = self._get(options['id'])
            if not xobject:
                raise HttpReqErrJson(404, "user not found. (id: %r)" % options['id'])

            xobject['servers']  = self.get_user_servers(options['id'])

            return xobject
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


    CREATE_SCHEMA   = xys.load("""
    username:           !~~regex(vpn_user.username)
    servers_selected:   !~~seqlen(1,255) [ !~~digit ]
    """)

    def create(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        args                = request.payload_params()
        if not xys.validate(args, self.CREATE_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            data                = dict(args)
            data['id']          = None
            data['enabled']     = True
            data['created_at']  = time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime())
            data['updated_at']  = data['created_at']

            servers             = data.pop('servers_selected')

            r                   = self._get(self.objects['vpn_user'].create(data.keys(), data.values()))

            for x in servers:
                server = self.objects['vpn_server'].get(**{'id': x, 'user_id': None})
                if not server:
                    LOG.error("unable to find server. (id: %r)", x)
                    continue

                if not self.objects['vpn_server'].update(('user_id',),
                                                         (r['id'],),
                                                         **{'id': server['id']}):
                    LOG.error("unable to update server. (id: %r)", server['id'])

                if not int(server['enabled']):
                    self.plugin_iptables.add_drop_sport('SERVER', server['port_number'])
                    self.plugin_iptables.add_drop_dport('SERVER', server['port_number'])
                else:
                    self.plugin_iptables.delete_drop_sport('SERVER', server['port_number'])
                    self.plugin_iptables.delete_drop_dport('SERVER', server['port_number'])

                self.plugin_openvpn.create_server_config(
                    self.plugin_openvpn.build_server_config(server))

                self.plugin_openvpn.cmd_call('enable', server['reference'])
                self.plugin_openvpn.cmd_call('restart', server['reference'])

            r['servers']        = self.get_user_servers(r['id'])

            return r
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def _update(self, xid, data):
        data['updated_at'] = time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime())

        return self.objects['vpn_user'].update(data.keys(),
                                               data.values(),
                                               **{'id': xid})


    UPDATE_SCHEMA   = xys.load("""
    id:                 !~~digit
    username:           !~~regex(vpn_user.username)
    servers_selected:   !~~seqlen(0,255) [ !~~digit ]
    """)

    def update(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        args = request.payload_params()
        if not xys.validate(args, self.UPDATE_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            xid                 = args.pop('id')
            xobject             = self._get(xid)

            if not xobject:
                raise HttpReqErrJson(404, "unable to find user. (id: %r)" % xid)

            args['username']    = args['username'].lower()

            servers             = args.pop('servers_selected')

            if not self._update(xid, args):
                raise HttpReqErrJson(500, "unable to update user. (id: %r)" % xid)

            servers_existing    = self.objects['vpn_server'].get_all(**{'user_id': xid})
            if servers_existing:
                for server in servers_existing:
                    if server['id'] in servers:
                        servers.remove(server['id'])
                    else:
                        self.objects['vpn_server'].update(('user_id',),
                                                          (None,),
                                                          **{'id': server['id']})

                        self.plugin_iptables.add_drop_sport('SERVER', server['port_number'])
                        self.plugin_iptables.add_drop_dport('SERVER', server['port_number'])

                        self.plugin_openvpn.cmd_call('stop', server['reference'])
                        self.plugin_openvpn.cmd_call('disable', server['reference'])
                        self.plugin_openvpn.remove_server_config(server['reference'])

            for x in servers:
                server = self.objects['vpn_server'].get(**{'id': x, 'user_id': None})
                if not server:
                    LOG.error("unable to find server. (id: %r)", x)
                    continue

                self.objects['vpn_server'].update(('user_id',),
                                                  (xid,),
                                                  **{'id': server['id']})

                if not int(server['enabled']):
                    self.plugin_iptables.add_drop_sport('SERVER', server['port_number'])
                    self.plugin_iptables.add_drop_dport('SERVER', server['port_number'])
                else:
                    self.plugin_iptables.delete_drop_sport('SERVER', server['port_number'])
                    self.plugin_iptables.delete_drop_dport('SERVER', server['port_number'])

                self.plugin_openvpn.create_server_config(
                    self.plugin_openvpn.build_server_config(server))

                self.plugin_openvpn.cmd_call('enable', server['reference'])
                self.plugin_openvpn.cmd_call('restart', server['reference'])

            r                   = self._get(xobject['id'])
            r['servers']        = self.get_user_servers(r['id'])

            return r
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


    _TOGGLE_ENABLED_SCHEMA  = xys.load("""
    id: !~~digit
    """)

    def _toggle_enabled(self, request, enabled):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not xys.validate(options, self._TOGGLE_ENABLED_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            xid     = options.pop('id')

            self._update(xid, {'enabled': helpers.boolize(enabled)})

            servers = self.objects['vpn_user'].get_servers(**{'vpn_user.id': xid})

            if servers:
                for server in servers:
                    if not int(server['vpn_user.enabled']):
                        self.plugin_iptables.add_drop_sport('USER', server['vpn_server.port_number'])
                        self.plugin_iptables.add_drop_dport('USER', server['vpn_server.port_number'])
                    else:
                        self.plugin_iptables.delete_drop_sport('USER', server['vpn_server.port_number'])
                        self.plugin_iptables.delete_drop_dport('USER', server['vpn_server.port_number'])

            xobject = self._get(xid)
            if not xobject:
                raise HttpReqErrJson(404, "unable to find user. (id: %r)" % xid)

            return xobject
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def enable(self, request):
        return self._toggle_enabled(request, True)

    def disable(self, request):
        return self._toggle_enabled(request, False)


    DELETE_SCHEMA  = xys.load("""
    id: !~~digit
    """)

    def delete(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options     = request.query_params()
        if not xys.validate(options, self.DELETE_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            xid     = options.pop('id')

            if not self.objects['vpn_user'].exists(**{'id': xid}):
                raise HttpReqErrJson(404, "unable to find user. (id: %r)" % xid)

            servers = self.objects['vpn_user'].get_servers(**{'vpn_user.id': xid})

            if servers:
                for server in servers:
                    self.plugin_iptables.add_drop_sport('USER', server['vpn_server.port_number'])
                    self.plugin_iptables.add_drop_dport('USER', server['vpn_server.port_number'])

                    self.plugin_openvpn.cmd_call('stop', server['vpn_server.reference'])
                    self.plugin_openvpn.cmd_call('disable', server['vpn_server.reference'])
                    self.plugin_openvpn.remove_server_config(server['vpn_server.reference'])

                    self.objects['vpn_server'].update(('user_id', 'updated_at'),
                                                      (None, time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime())),
                                                      **{'id': server['vpn_server.id']})

            self.objects['vpn_user'].delete(**{'id': xid})
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


    VALIDATE_SCHEMA = xys.load("""
    username:   !~~regex(vpn_user.username)
    """)

    def validate(self, request):
        self.plugin_credentials.set_request(request)
        if not self.plugin_credentials.is_admin():
            raise HttpReqErrJson(403, "bad credentials")

        options = request.query_params()
        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            for key, value in iteritems(options):
                if not self.VALIDATE_SCHEMA.has_key(key):
                    raise HttpReqErrJson(404, "unknown argument for command. (key: %r)" % key)
                if not xys.validate({key: value}, {key: self.VALIDATE_SCHEMA[key]}):
                    raise HttpReqErrJson(415, "invalid argument for command. (key: %r)" % key)

            return True
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

if __name__ != "__main__":
    def _start():
        MODULES.register(VPNUserModule())
    _start()
