# -*- coding: utf-8 -*-
"""iptables plugin"""

import logging
import iptc

from dwho.classes.plugins import DWhoPluginBase, PLUGINS


LOG         = logging.getLogger('fd-vpn.plugins.iptables')
CHAIN_NAME  = 'FDVPN'
SYS_CHAINS  = ('INPUT', 'FORWARD', 'OUTPUT')
USER_CHAINS = ('CLIENT', 'SERVER', 'USER')


class FdVpnIptablesPlugin(DWhoPluginBase):
    PLUGIN_NAME = 'iptables'

    @staticmethod
    def _get_chain_name(chain):
        return "%s_%s" % (CHAIN_NAME, chain)

    def _build_chain(self, sys_chain, user_chain):
        chain_name      = "%s_%s" % (CHAIN_NAME, user_chain)

        if self.table.is_chain(chain_name):
            self.users_chain[chain_name] = iptc.Chain(self.table, chain_name)
        else:
            self.users_chain[chain_name] = self.table.create_chain(chain_name)

        rule            = iptc.Rule()
        rule.create_target(chain_name)

        attr = getattr(self, sys_chain)
        if attr:
            try:
                attr.delete_rule(rule)
            except iptc.ip4tc.IPTCError:
                attr.insert_rule(rule)

        """
        for xsys_chain in SYS_CHAINS:
            while True:
                attr = getattr(self, xsys_chain)
                try:
                    attr.delete_rule(rule)
                except iptc.ip4tc.IPTCError:
                    attr.insert_rule(rule)
                    break
        """

    def at_start(self):
        self.table          = iptc.Table(iptc.Table.FILTER)
        self.users_chain    = {}

        for sys_chain in SYS_CHAINS:
            setattr(self, sys_chain, iptc.Chain(self.table, sys_chain))

        for sys_chain in SYS_CHAINS:
            for user_chain in USER_CHAINS:
                self._build_chain(sys_chain, user_chain)

    def _add_drop_cidr(self, xtype, chain, cidr):
        chain       = self._get_chain_name(chain)
        rule        = iptc.Rule()
        setattr(rule, xtype, cidr)
        target      = iptc.Target(rule, 'DROP')
        rule.target = target

        while True:
            try:
                self.users_chain[chain].delete_rule(rule)
            except iptc.ip4tc.IPTCError:
                self.users_chain[chain].insert_rule(rule)
                break

    def add_drop_src(self, chain, cidr):
        self._add_drop_cidr('src', chain, cidr)

    def add_drop_dst(self, chain, cidr):
        self._add_drop_cidr('dst', chain, cidr)

    def _add_drop_port(self, xtype, chain, port):
        chain           = self._get_chain_name(chain)
        rule            = iptc.Rule()
        rule.protocol   = 'udp'
        match           = rule.create_match('udp')
        setattr(match, xtype, str(port))
        target          = iptc.Target(rule, 'DROP')
        rule.target     = target

        while True:
            try:
                self.users_chain[chain].delete_rule(rule)
            except iptc.ip4tc.IPTCError:
                self.users_chain[chain].insert_rule(rule)
                break

    def add_drop_dport(self, chain, port):
        self._add_drop_port('dport', chain, port)

    def add_drop_sport(self, chain, port):
        self._add_drop_port('sport', chain, port)

    def _delete_drop_cidr(self, xtype, chain, cidr):
        chain       = self._get_chain_name(chain)
        rule        = iptc.Rule()
        setattr(rule, xtype, cidr)
        target      = iptc.Target(rule, 'DROP')
        rule.target = target

        while True:
            try:
                self.users_chain[chain].delete_rule(rule)
            except iptc.ip4tc.IPTCError:
                break

    def delete_drop_src(self, chain, cidr):
        self._delete_drop_cidr('src', chain, cidr)

    def delete_drop_dst(self, chain, cidr):
        self._delete_drop_cidr('dst', chain, cidr)

    def _delete_drop_port(self, xtype, chain, port):
        chain           = self._get_chain_name(chain)
        rule            = iptc.Rule()
        rule.protocol   = 'udp'
        match           = rule.create_match('udp')
        setattr(match, xtype, str(port))
        target          = iptc.Target(rule, 'DROP')
        rule.target     = target

        while True:
            try:
                self.users_chain[chain].delete_rule(rule)
            except iptc.ip4tc.IPTCError:
                break

    def delete_drop_dport(self, chain, port):
        self._delete_drop_port('dport', chain, port)

    def delete_drop_sport(self, chain, port):
        self._delete_drop_port('sport', chain, port)


if __name__ != "__main__":
    def _start():
        PLUGINS.register(FdVpnIptablesPlugin())
    _start()
