# -*- coding: utf-8 -*-
"""openvpn plugin"""

import glob
import grp
import logging
import os
import pwd
import re
import shlex
import shutil
import socket
import subprocess
import time

from six.moves.configparser import ConfigParser
from six import StringIO

from fd_vpn.objects.vpn_server import FdVpnObjectVPNServer
from fd_vpn.objects.vpn_user import FdVpnObjectVPNUser

from dwho.classes.plugins import DWhoPluginBase, PLUGINS
from sonicprobe import helpers
from sonicprobe.libs import network
from sonicprobe.libs.openvpn import OpenVPNMgmt


LOG             = logging.getLogger('fd-vpn.plugins.openvpn')
CFG_OPT_DIRS    = {'file':  ('ifconfig-pool-persist',
                             'log',
                             'log-append',
                             'status',
                             'tls-auth'),
                   'folder': ('client-config-dir',)}
RE_NAME_VALUE   = re.compile(r'\s*([^\s]+)\s*(.*)$').match


class FdVpnOpenVpnPlugin(DWhoPluginBase):
    PLUGIN_NAME = 'openvpn'

    @staticmethod
    def _generate_ta_key(cmd_gen_ta_key, filepath):
        LOG.debug("generate ta.key file")

        cmd         = subprocess.Popen(shlex.split(cmd_gen_ta_key),
                                       stderr       = subprocess.PIPE,
                                       close_fds    = True)
        (out, err)  = cmd.communicate()

        if cmd.returncode:
            raise RuntimeError("unable to create %r. (err: %r, rc: %r)"
                                % (filepath.split(' ', 1)[0],
                                   err,
                                   cmd.returncode))

    def at_start(self):
        return

    def cmd_call(self, action, name = None):
        LOG.debug("call command: %r, name: %r", action, name)

        if action not in self.config['vpn']['server']['commands']:
            LOG.warning("command unknown. (command: %r, name: %r)", action, name)
            return

        cmd         = subprocess.Popen(
                          shlex.split(self.config['vpn']['server']['commands'][action] % {'name': name}),
                          stderr       = subprocess.PIPE,
                          close_fds    = True)
        (out, err)  = cmd.communicate()

        if cmd.returncode:
            raise RuntimeError("unable to %r. (name: %r, err: %r, rc: %r)"
                               % (action, name, err, cmd.returncode))

    def build_server_config(self, xobject):
        cidr = network.parse_ipv4_cidr(xobject['cidr'])
        if not cidr:
            raise ValueError("invalid cidr. (cidr: %r)" % cidr)

        return {'server_address':       self.config['vpn']['server']['address'],
                'server_reference':     xobject['reference'],
                'server_port_number':   xobject['port_number'],
                'server_ip_addr':       cidr[0],
                'server_ip_subnet':     network.bitmask_to_netmask_ipv4(cidr[1])}

    def get_server_config(self, server_reference, xobject = None):
        config = dict(self.config['vpn']['server'])
        config.update({'server_reference': server_reference})

        if isinstance(xobject, dict):
            config.update(xobject)

        server_conf_file = self.config['vpn']['server']['conf_file'] % config

        if not os.path.isfile(server_conf_file):
            LOG.warning("unable to find server config file. (file: %r)",
                        server_conf_file)
            return (server_conf_file, config)

        with open(server_conf_file, 'r') as f:
            lines = ["[general]"]

            for line in f.readlines():
                m = RE_NAME_VALUE(line)
                if m:
                    name = m.group(1).strip()
                    value = m.group(2).strip()

                    if name == 'push':
                        lines.append("%s \"%s\" = yes" % (name, value))
                    else:
                        lines.append("%s = %s" % (name, value))

        config = ConfigParser()
        config.readfp(StringIO("\r\n".join(lines) + "\r\n"))

        return (server_conf_file, config)

    def create_server_config(self, cfg, generate_ta_key = False):
        cfg.update(self.config['vpn']['server'])

        server_conf_file    = self.config['vpn']['server']['conf_file'] % cfg
        cmd_gen_ta_key      = self.config['vpn']['server']['commands']['gen_ta_key'] % cfg

        with open(self.config['vpn']['server']['conf_tpl'], 'r') as f:
            lines = StringIO('\r\n'.join(helpers.txtsubst(f.readlines(), cfg, charset = 'utf8')))

        config              = ConfigParser()
        config.readfp(lines)

        lines = []

        for option in sorted(config.options('general')):
            if config.get('general', option) != 'yes':
                lines.append("%s %s" % (option, config.get('general', option)))
            else:
                lines.append("%s" % option)

        helpers.file_w_tmp("\r\n".join(lines) + "\r\n", path = server_conf_file)

        uid = pwd.getpwnam(config.get('general', 'user'))[2]
        gid = grp.getgrnam(config.get('general', 'group'))[2]

        for cfg_opt_dir_type, cfg_opt_names in CFG_OPT_DIRS.iteritems():
            for cfg_opt_name in cfg_opt_names:
                pathname = config.get('general', cfg_opt_name)
                if cfg_opt_dir_type == 'file':
                    dirname = os.path.dirname(pathname)
                else:
                    dirname = pathname

                if not os.path.isdir(dirname):
                    os.makedirs(dirname, mode = 0700)

                if os.path.exists(dirname):
                    os.chown(dirname, uid, gid)

                if os.path.exists(pathname):
                    os.chown(pathname, uid, gid)

                if os.path.isfile(pathname):
                    os.chmod(pathname, 0600)

                if os.path.isdir(pathname):
                    os.chmod(pathname, 0700)

        tls_auth_opt = config.get('general', 'tls-auth')

        if not os.path.isfile(tls_auth_opt.split(' ', 1)[0]) or generate_ta_key:
            self._generate_ta_key(cmd_gen_ta_key, tls_auth_opt)

    def remove_server_config(self, server_reference):
        (server_conf_file,
         config) = self.get_server_config(server_reference)

        client_config_dir = config.get('general', 'client-config-dir')

        if os.path.isdir(client_config_dir):
            shutil.rmtree(client_config_dir)

        tls_auth_file = config.get('general', 'tls-auth').split(' ', 1)[0]
        if os.path.isfile(tls_auth_file):
            os.unlink(tls_auth_file)

        ifconfig_pool_persist_file = config.get('general', 'ifconfig-pool-persist')
        if os.path.isfile(ifconfig_pool_persist_file):
            os.unlink(ifconfig_pool_persist_file)

        if os.path.isfile(server_conf_file):
            os.unlink(server_conf_file)

        dirname  = os.path.dirname(config.get('general', 'client-config-dir'))
        if os.path.isdir(dirname):
            os.rmdir(dirname)

    def get_client_config(self, client_reference, client_auth_method, server_object):
        cfg = dict(self.config['vpn']['server'])
        cfg.update(self.build_server_config(server_object))
        cfg['client_reference'] = client_reference

        if client_auth_method in ('otp', 'userpass'):
            cfg['client_auth_method'] = 'auth-user-pass'
        else:
            cfg['client_auth_method'] = '#auth-user-pass'

        with open(self.config['vpn']['client']['conf_tpl'], 'r') as f:
            return StringIO('\r\n'.join(helpers.txtsubst(f.readlines(), cfg, charset = 'utf8')))

    def get_client_config_inline(self, client_reference, client_auth_method, server_object, files):
        cfg = dict(self.config['vpn']['server'])
        cfg.update(self.build_server_config(server_object))
        cfg['client_reference'] = client_reference

        if client_auth_method in ('otp', 'userpass'):
            cfg['client_auth_method'] = 'auth-user-pass'
        else:
            cfg['client_auth_method'] = '#auth-user-pass'

        for key, xfile in files.iteritems():
            if os.path.isfile(xfile):
                with open(xfile, 'r') as f:
                    cfg[key] = f.read()

        with open(self.config['vpn']['client']['conf_inline_tpl'], 'r') as f:
            return StringIO('\r\n'.join(helpers.txtsubst(f.readlines(), cfg, charset = 'utf8')))

    def create_ccd_config(self, server_reference, server_cidr, xobject):
        (server_conf_file,
         config)            = self.get_server_config(server_reference)

        if not os.path.isdir(config.get('general', 'client-config-dir')):
            LOG.warning("unable to find client config dir. (dir: %r)",
                        config.get('general', 'client-config-dir'))
            return

        cidr = network.parse_ipv4_cidr(server_cidr)
        if not cidr:
            raise ValueError("invalid cidr. (cidr: %r)" % cidr)

        addr_ip             = cidr[0].split('.')[0:3]

        client_ip           = '.'.join(addr_ip + [str(xobject['last_octet'])])

        client_conf_file    = os.path.join(config.get('general', 'client-config-dir'),
                                           (self.config['vpn']['client']['common_name']
                                            % {'vpn_client.reference': xobject['reference'],
                                               'vpn_server.reference': server_reference}))

        helpers.file_w_tmp("ifconfig-push %s %s\r\n" % (client_ip,
                                                        network.bitmask_to_netmask_ipv4(cidr[1])),
                           client_conf_file,
                           mode = 'w+')

        uid = pwd.getpwnam(config.get('general', 'user'))[2]
        gid = grp.getgrnam(config.get('general', 'group'))[2]

        if os.path.exists(client_conf_file):
            os.chown(client_conf_file, uid, gid)

        if os.path.isfile(client_conf_file):
            os.chmod(client_conf_file, 0600)

    def _toggle_ccd_config(self, server_reference, xobject):
        (server_conf_file,
         config)            = self.get_server_config(server_reference)

        if not os.path.isdir(config.get('general', 'client-config-dir')):
            LOG.warning("unable to find client config dir. (dir: %r)",
                        config.get('general', 'client-config-dir'))
            return

        client_conf_file    = os.path.join(config.get('general', 'client-config-dir'),
                                           (self.config['vpn']['client']['common_name']
                                            % {'vpn_client.reference': xobject['reference'],
                                               'vpn_server.reference': server_reference}))

        if not os.access(client_conf_file, os.R_OK | os.W_OK):
            LOG.warning("unable to find client file. (file: %r)",
                        client_conf_file)
            return

        return (server_conf_file,
                config,
                client_conf_file)

    def enable_ccd_config(self, server_reference, xobject):
        configs     = self._toggle_ccd_config(server_reference, xobject)
        if not configs:
            return

        lines       = []
        disabled    = False

        with open(configs[2], 'r') as f:
            for line in f.readlines():
                if line.strip() != 'disable':
                    lines.append(line.strip())
                else:
                    disabled = True

        if disabled:
            helpers.file_w_tmp("\r\n".join(lines) + "\r\n", configs[2], mode = 'w+')

    def disable_ccd_config(self, server_reference, xobject):
        configs = self._toggle_ccd_config(server_reference, xobject)
        if not configs:
            return

        lines   = []

        with open(configs[2], 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if line == 'disable':
                    return
                lines.append(line)

        lines.append('disable')

        helpers.file_w_tmp("\r\n".join(lines) + "\r\n", configs[2], mode = 'w+')

        (mgmt_host,
         mgmt_port) = configs[1].get('general', 'management').split(' ', 1)

        mgmt    = OpenVPNMgmt(mgmt_host, int(mgmt_port), start_open = True)
        mgmt.kill(self.config['vpn']['client']['common_name']
                  % {'vpn_client.reference': xobject['reference'],
                     'vpn_server.reference': server_reference})
        mgmt.close()

    def remove_ccd_config(self, server_reference, xobject):
        (server_conf_file,
         config)            = self.get_server_config(server_reference)

        if not os.path.isdir(config.get('general', 'client-config-dir')):
            LOG.warning("unable to find client config dir. (dir: %r)",
                        config.get('general', 'client-config-dir'))
            return

        client_conf_file    = os.path.join(config.get('general', 'client-config-dir'),
                                           xobject['reference'])

        for x in glob.iglob("%s.*" % client_conf_file):
            os.unlink(x)

        (mgmt_host,
         mgmt_port) = config.get('general', 'management').split(' ', 1)

        mgmt    = OpenVPNMgmt(mgmt_host, int(mgmt_port), start_open = True)
        mgmt.kill(self.config['vpn']['client']['common_name']
                  % {'vpn_client.reference': xobject['reference'],
                     'vpn_server.reference': server_reference})
        mgmt.close()


if __name__ != "__main__":
    def _start():
        PLUGINS.register(FdVpnOpenVpnPlugin())
    _start()
